/******************************************************************************
*
* project name:    TI-Chess / BookBuilder
* file name:       bin2oth.c
* initial date:    09/10/2002
* author:          thomas.nussbaumer@gmx.net
* description:     parses input file and generates position list
*
* (taken from TIGCC Tools Suite and modified to fit our needs)
*
* $Id: bin2oth.c,v 1.4 2002/10/23 21:38:52 tnussb Exp $
*
******************************************************************************/

typedef struct {
    char          signature[8]; // "**TI92P*" or "**TI89**"
    unsigned char fill1[2];     // 01 00
    char          folder[8];    // folder name
    char          desc[40];     // ---- not used ----
    unsigned char fill2[6];     // 01 00 52 00 00 00
    char          name[8];      // varname
    unsigned char type[4];      // 0C 00 00 00
    unsigned char size[4];      // complete file size (including checksum)
    unsigned char fill3[6];     // A5 5A 00 00 00 00
    unsigned char datasize[2];  // data size
}
StrHeader;


//=============================================================================
// returns a pointer to a newly allocated memory block which holds the
// OTH file contents ready to flush into a file (returns 0 if error)
//
// the length of the newly allocated memory block will be stored in outlength
//
// NOTE: the length of the extension may vary between 1 and 4 characters
//       (no checks are performed on the type of the character)
//=============================================================================
unsigned char* DataBuffer2OTHBuffer(int            calctype,
                                    char*          folder,
                                    char*          varname,
                                    char*          extension,
                                    unsigned int   inlength,
                                    unsigned char* indata,
                                    unsigned int*  outlength)
{
    StrHeader*     psh;
    unsigned int   checksum;
    unsigned char* ptr;
    unsigned int   size;
    unsigned char* output;
    int            ext_len;

    if (!extension) {
        fprintf(stderr,"ERROR: no extension given\n");
        return 0;
    }

    ext_len = strlen(extension);
    if (ext_len < 1 || ext_len > 4) {
        fprintf(stderr,"ERROR: invalid OTH extension [%s]\n",extension);
        return 0;
    }


    if (inlength > TT_MAX_OTHDATA) {
        fprintf(stderr,"ERROR: length (%u) exceeds max. user data size (%u)\n",inlength,TT_MAX_OTHDATA);
        return 0;
    }

    size = sizeof(StrHeader)+inlength+2+3+ext_len; // 6 == OTH bytes , 2 == checksum

    if (!(output = (unsigned char*)malloc(size))) {
        fprintf(stderr,"ERROR: cannot allocate %u bytes of memory\n",size);
        return 0;
    }

    *outlength = size;
    psh = (StrHeader*)output;
    memset(psh,0,sizeof(StrHeader));

    //-------------------------------------------------------------------
    // fill up all the static fields
    //-------------------------------------------------------------------
    psh->fill1[0]=1;psh->fill1[1]=0;
    psh->fill2[0]=0x01;psh->fill2[1]=0x00;psh->fill2[2]=0x52;
    psh->fill2[3]=0x00;psh->fill2[4]=0x00;psh->fill2[5]=0x00;
    psh->fill3[0]=0xA5;psh->fill3[1]=0x5A;psh->fill3[2]=0x00;
    psh->fill3[3]=0x00;psh->fill3[4]=0x00;psh->fill3[5]=0x00;
    psh->type[0]=0x1c;psh->type[1]=0x00;psh->type[2]=0x00;psh->type[3]=0x00;

    //-------------------------------------------------------------------
    // fill in the magic marker string depending on given calc type
    //-------------------------------------------------------------------
    if (calctype == CALC_TI89) strncpy(psh->signature,SIGNATURE_TI89,8);
    else                       strncpy(psh->signature,SIGNATURE_TI92P,8);

    //-------------------------------------------------------------------
    // fill in folder and variable name
    // if folder name pointer is NULL, use DEFAULT_FOLDER ("main")
    //-------------------------------------------------------------------
    if (!folder) strncpy(psh->folder,DEFAULT_FOLDER,8);
    else         strncpy(psh->folder,folder,8);

    strncpy(psh->name,varname,8);

    //-------------------------------------------------------------------
    // size holds the complete output size == filelength including header
    //-------------------------------------------------------------------
    psh->size[0] = (unsigned char)(size & 0xff);
    psh->size[1] = (unsigned char)((size >> 8)  & 0xff);
    psh->size[2] = (unsigned char)((size >> 16) & 0xff);
    psh->size[3] = (unsigned char)((size >> 24) & 0xff);

    //-------------------------------------------------------------------
    // data size will hold user data size + OTH tag related bytes
    //-------------------------------------------------------------------
    size -= sizeof(StrHeader);
    size -= 2;
    psh->datasize[0] = (unsigned char)((size >> 8) & 0xff);
    psh->datasize[1] = (unsigned char)(size & 0xff);

    //-------------------------------------------------------------------
    // copy complete indata
    //-------------------------------------------------------------------
    memcpy(output + sizeof(StrHeader),indata,inlength);

    //-------------------------------------------------------------------
    // append OTH bytes
    //-------------------------------------------------------------------
    ptr    = output + sizeof(StrHeader) + inlength;
    *ptr++ = 0;

    switch (ext_len) {
        case 1:
            *ptr++ = *extension;
            *ptr++ = 0;
            break;
        case 2:
            *ptr++ = *extension++;
            *ptr++ = *extension;
            *ptr++ = 0;
            break;
        case 3:
            *ptr++ = *extension++;
            *ptr++ = *extension++;
            *ptr++ = *extension;
            *ptr++ = 0;
            break;
        case 4:
            *ptr++ = *extension++;
            *ptr++ = *extension++;
            *ptr++ = *extension++;
            *ptr++ = *extension;
            *ptr++ = 0;
            break;
    }
    *ptr++ = 0xF8;

    size = *outlength - sizeof(StrHeader);
    ptr  = psh->datasize;

    checksum = 0;
    while (size--) checksum += *ptr++;

    output[*outlength-2] = (unsigned char)(checksum & 0xff);
    output[*outlength-1] = (unsigned char)((checksum >> 8) & 0xff);

    return output;
}

#if defined(BOOK_COMPILER)
//=============================================================================
// generates TI89 and TI92p/V200 files from given inputbuffer and varname
//
// the globals param_noti89 and param_ti92p will control which files are
// generated
//=============================================================================
int GenerateOutputFiles(char* varname,unsigned char* inbuffer,unsigned int inlen,int side) {
    FILE*          fp;
    char           filename[255];
    unsigned char* othdata;
    unsigned int   othdata_len;
    char           tmpstr[255];
    short          file_generated = 0;

    if (!param_noti89) {
        //---------------------------------------------------------------------
        // generate and write TI89 file
        //---------------------------------------------------------------------
        if (!(othdata = DataBuffer2OTHBuffer(CALC_TI89,
                                             param_folder,
                                             varname,
                                             "tict",
                                             inlen,
                                             inbuffer,
                                             &othdata_len)))
        {
            WriteLog("ERROR: othdata generation for TI89 failed\n",MIRROR_TO_STDERR);
            return 0;
        }

        sprintf(filename,"%s.89y",varname);

        if (!(fp = fopen(filename,"wb"))) {
            sprintf(tmpstr,"ERROR: cannot open TI89 file %s\n",filename);
            WriteLog(tmpstr,MIRROR_TO_STDERR);
            free(othdata);
            return 0;
        }

        if (fwrite(othdata,othdata_len,1,fp)!=1) {
            sprintf(tmpstr,"ERROR: cannot write to TI89 file %s\n",filename);
            WriteLog(tmpstr,MIRROR_TO_STDERR);
            fclose(fp);
            free(othdata);
            return 0;
        }

        sprintf(tmpstr,"%s book %s for TI89 (oncalcsize=%d) generated\n",
                ((side==WHITE)? "WHITE":"BLACK"),
                filename,
                inlen+9);
        WriteLog(tmpstr,MIRROR_TO_STDERR);
        file_generated = 1;
        free(othdata);
    }


    if (!param_noti92p) {
        //---------------------------------------------------------------------
        // generate and write TI92p/V200 file
        //---------------------------------------------------------------------
        if (!(othdata = DataBuffer2OTHBuffer(CALC_TI92P,
                                             param_folder,
                                             varname,
                                             "tict",
                                             inlen,
                                             inbuffer,
                                             &othdata_len)))
        {
            WriteLog("ERROR: othdata generation for TI89 failed\n",MIRROR_TO_STDERR);
            return 0;
        }

        sprintf(filename,"%s.9xy",varname);

        if (!(fp = fopen(filename,"wb"))) {
            sprintf(tmpstr,"ERROR: cannot open TI92p/V200 file %s\n",filename);
            WriteLog(tmpstr,MIRROR_TO_STDERR);
            free(othdata);
            return 0;
        }

        if (fwrite(othdata,othdata_len,1,fp)!=1) {
            sprintf(tmpstr,"ERROR: cannot write to TI92p/V200 file %s\n",filename);
            WriteLog(tmpstr,MIRROR_TO_STDERR);
            fclose(fp);
            free(othdata);
            return 0;
        }

        sprintf(tmpstr,"%s book %s for TI92p/V200 (oncalcsize=%d) generated\n",
                ((side==WHITE)? "WHITE":"BLACK"),
                filename,
                inlen+9);
        WriteLog(tmpstr,MIRROR_TO_STDERR);

        file_generated = 1;
        free(othdata);
    }

    if (file_generated) {
        param_offset++;
    }

    return 1;
}
#endif


//#############################################################################
//###################### NO MORE FAKES BEYOND THIS LINE #######################
//#############################################################################
//
//=============================================================================
// Revision History
//=============================================================================
//
// $Log: bin2oth.c,v $
// Revision 1.4  2002/10/23 21:38:52  tnussb
// file index counting fixed
//
// Revision 1.3  2002/10/18 15:16:45  tnussb
// changes due to -name and -folder commandline parameters
//
// Revision 1.2  2002/10/10 13:06:26  tnussb
// minor modifications, because this file is shared between cpc and bkc yet
//
// Revision 1.1  2002/10/09 15:49:23  tnussb
// initial version
//
//
